import sys
import types
import pytest


PYTEST_GTE_7_0 = getattr(pytest, 'version_tuple', (0, 0)) >= (7, 0)
PYTEST_GTE_5_4 = PYTEST_GTE_7_0 or hasattr(pytest.Collector, 'from_parent')


def trace_function(funcobj, *args, **kwargs):
    """Call a function, and return its locals"""
    funclocals = {}

    def _tracefunc(frame, event, arg):
        # Activate local trace for first call only
        if frame.f_back.f_locals.get('_tracefunc') == _tracefunc:
            if event == 'return':
                funclocals.update(frame.f_locals)

    sys.setprofile(_tracefunc)
    try:
        funcobj(*args, **kwargs)
    finally:
        sys.setprofile(None)

    return funclocals


def make_module_from_function(funcobj):
    """Evaluates the local scope of a function, as if it was a module"""
    module = types.ModuleType(funcobj.__name__)

    # Import shared behaviors into the generated module. We do this before
    # importing the direct children, so that fixtures in the block that's
    # importing the behavior take precedence.
    for shared_funcobj in getattr(funcobj, '_behaves_like', []):
        module.__dict__.update(evaluate_shared_behavior(shared_funcobj))

    # Import children
    module.__dict__.update(trace_function(funcobj))
    return module


def evaluate_shared_behavior(funcobj):
    if not hasattr(funcobj, '_shared_functions'):
        funcobj._shared_functions = {}
        for name, obj in trace_function(funcobj).items():
            # Only functions are relevant here
            if not isinstance(obj, types.FunctionType):
                continue

            # Mangle names of imported functions, except fixtures because we
            # want fixtures to be overridden in the block that's importing the
            # behavior.
            if not hasattr(obj, '_pytestfixturefunction'):
                name = obj._mangled_name = f"{funcobj.__name__}::{name}"

            funcobj._shared_functions[name] = obj
    return funcobj._shared_functions


class DescribeBlock(pytest.Module):
    """Module-like object representing the scope of a describe block"""

    @classmethod
    def from_parent(cls, parent, obj):
        """Construct a new node for the describe block"""
        name = getattr(obj, '_mangled_name', obj.__name__)
        nodeid = parent.nodeid + '::' + name
        if PYTEST_GTE_7_0:
            self = super().from_parent(
                parent=parent, path=parent.path, nodeid=nodeid)
        elif PYTEST_GTE_5_4:
            self = super().from_parent(
                parent=parent, fspath=parent.fspath, nodeid=nodeid)
        else:
            self = cls(parent=parent, fspath=parent.fspath, nodeid=nodeid)
        self.name = name
        self.funcobj = obj
        return self

    def collect(self):
        """Get list of children"""
        self.session._fixturemanager.parsefactories(self)
        return super().collect()

    def _getobj(self):
        """Get the underlying Python object"""
        return self._importtestmodule()

    def _importtestmodule(self):
        """Import a describe block as if it was a module"""
        module = make_module_from_function(self.funcobj)
        self.own_markers = getattr(self.funcobj, 'pytestmark', [])
        return module

    def funcnamefilter(self, name):
        """Treat all nested functions as tests

        We do not require the 'test_' prefix for the specs.
        """
        return not name.startswith('_')

    def classnamefilter(self, name):
        """Don't allow test classes inside describe"""
        return False

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


def pytest_pycollect_makeitem(collector, name, obj):
    if isinstance(obj, types.FunctionType):
        for prefix in collector.config.getini('describe_prefixes'):
            if obj.__name__.startswith(prefix):
                return DescribeBlock.from_parent(collector, obj)


def pytest_addoption(parser):
    parser.addini("describe_prefixes", type="args", default=("describe",),
                  help="prefixes for Python describe function discovery")



def behaves_like(*behavior_funcs):
    def decorator(func):
        if not hasattr(func, '_behaves_like'):
            func._behaves_like = []
        func._behaves_like += behavior_funcs
        return func
    return decorator



import py
import re

pytest_plugins = 'pytester'

ini = """
[pytest]
describe_prefixes = foo bar
"""


def test_collect_custom_prefix(testdir):
    testdir.makeini(ini)

    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        def foo_scope():
            def bar_context():
                def passes():
                    pass
    """))

    result = testdir.runpytest('--collectonly')
    expected_lines = map(re.compile, [
        r"collected 1 item(s)?",
        r"\s*<Module '?(a_dir/)?test_a.py'?>",
        r"\s*<DescribeBlock '?foo_scope'?>",
        r"\s*<DescribeBlock '?bar_context'?>",
        r"\s*<Function '?passes'?>",
    ])
    for line in expected_lines:
        assert any([line.match(r) is not None for r in result.outlines])



import py
from util import assert_outcomes

pytest_plugins = 'pytester'


def test_special_marks(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest

        def describe_marks():
            @pytest.mark.xfail
            def xfails():
                assert False

            @pytest.mark.xfail
            def xpasses():
                pass

            @pytest.mark.skipif("0 < 1")
            def skipped():
                pass

            @pytest.mark.parametrize('foo', (1, 2, 3))
            def isint(foo):
                assert foo == int(foo)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=3, xfailed=1, xpassed=1, skipped=1)


def test_multiple_variables_parametrize(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest

        def describe_marks():
            @pytest.mark.parametrize('foo,bar', [(1, 2), (3, 4)])
            def isint_str_names(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)

            @pytest.mark.parametrize(['foo', 'bar'], [(1, 2), (3, 4)])
            def isint_list_names(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)

            @pytest.mark.parametrize(('foo', 'bar'), [(1, 2), (3, 4)])
            def isint_tuple_names(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=6)


def test_cartesian_parametrize(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest

        def describe_marks():

            @pytest.mark.parametrize('foo', (1, 2, 3))
            @pytest.mark.parametrize('bar', (1, 2, 3))
            def isint(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=9)


def test_parametrize_applies_to_describe(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest

        @pytest.mark.parametrize('foo', (1, 2, 3))
        def describe_marks():

            @pytest.mark.parametrize('bar', (1, 2, 3))
            def isint(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)

            def isint2(foo):
                assert foo == int(foo)

            def describe_nested():
                def isint3(foo):
                    assert foo == int(foo)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=15)


def test_cartesian_parametrize_on_describe(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest

        @pytest.mark.parametrize('foo', (1, 2, 3))
        @pytest.mark.parametrize('bar', (1, 2, 3))
        def describe_marks():

            def isint(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=9)


def test_parametrize_with_shared(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        from pytest import fixture
        from pytest_describe import behaves_like

        def a_duck():
            def it_quacks(sound):
                assert sound == int(sound)


        @pytest.mark.parametrize('foo', (1, 2, 3))
        @behaves_like(a_duck)
        def describe_something_that_quacks():
            @fixture
            def sound(foo):
                return foo

        @pytest.mark.parametrize('foo', (1, 2, 3))
        @behaves_like(a_duck)
        def describe_something_that_barks():
            @fixture
            def sound(foo):
                return foo
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=6)


def test_parametrize_with_shared_but_different_values(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        from pytest import fixture
        from pytest_describe import behaves_like

        def a_duck():
            def it_quacks(sound):
                assert sound[1] == int(sound[1])
                assert sound[0] == 'bark' or sound[1] <= 3
                assert sound[0] == 'quack' or sound[1] >= 4


        @pytest.mark.parametrize('foo', (1, 2, 3))
        @behaves_like(a_duck)
        def describe_something_that_quacks():
            @fixture
            def sound(foo):
                return ('quack', foo)

        @pytest.mark.parametrize('foo', (4, 5, 6))
        @behaves_like(a_duck)
        def describe_something_that_barks():
            @fixture
            def sound(foo):
                return ('bark', foo)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=6)


def test_coincident_parametrize_at_top(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest

        @pytest.mark.parametrize('foo', (1, 2, 3))
        def describe_marks():

            @pytest.mark.parametrize('bar', (1, 2, 3))
            def isint(foo, bar):
                assert foo == int(foo)
                assert bar == int(bar)

        @pytest.mark.parametrize('foo', (1, 2, 3))
        def describe_marks2():
            def isint2(foo):
                assert foo == int(foo)
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=12)


def test_keywords(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        def describe_a():
            @pytest.mark.foo
            def foo_test():
                pass
            @pytest.mark.bar
            def bar_test():
                pass
    """))

    result = testdir.runpytest('-k', 'foo')
    assert_outcomes(result, passed=1, deselected=1)


def test_marks(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        def describe_a():
            @pytest.mark.foo
            def foo_test():
                pass
            @pytest.mark.bar
            def bar_test():
                pass
    """))

    result = testdir.runpytest('-m', 'foo')
    assert_outcomes(result, passed=1, deselected=1)


def test_module_marks(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        pytestmark = [ pytest.mark.foo ]
        def describe_a():
            pytestmark = [ pytest.mark.bar ]
            def describe_b():
                def a_test():
                    pass
    """))

    result = testdir.runpytest('-m', 'foo')
    assert_outcomes(result, passed=1)


def test_mark_at_describe_function(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        @pytest.mark.foo
        def describe_foo():
            def describe_a():
                def a_test():
                    pass
            @pytest.mark.bar
            def b_test():
                pass
    """))

    result = testdir.runpytest('-m', 'foo')
    assert_outcomes(result, passed=2)


def test_mark_stacking(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        import pytest
        @pytest.fixture()
        def get_marks(request):
            return [(mark.args[0], node.name) for node, mark
                    in request.node.iter_markers_with_node(name='my_mark')]

        @pytest.mark.my_mark('foo')
        def describe_marks():
            def it_is_inherited_from_describe_block(get_marks):
                assert get_marks == [('foo', 'describe_marks')]

            @pytest.mark.my_mark('bar')
            @pytest.mark.my_mark('baz')
            def all_marks_are_chained(get_marks):
                assert get_marks == [
                    ('baz', 'all_marks_are_chained'),
                    ('bar', 'all_marks_are_chained'),
                    ('foo', 'describe_marks')]
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=2)



def assert_outcomes(result, **expected):
    outcomes = result.parseoutcomes()

    for key in 'seconds', 'warning', 'warnings':
        if key in outcomes:
            del outcomes[key]

    assert outcomes == expected



import py
from util import assert_outcomes

pytest_plugins = 'pytester'


def test_shared_behavior(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        def a_duck():
            def it_quacks(sound):
                assert sound == "quack"

        @behaves_like(a_duck)
        def describe_something_that_quacks():
            @fixture
            def sound():
                return "quack"

        @behaves_like(a_duck)
        def describe_something_that_barks():
            @fixture
            def sound():
                return "bark"
    """))

    result = testdir.runpytest()
    assert_outcomes(result, failed=1, passed=1)


def test_multiple_shared_behaviors(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        def a_duck():
            def it_quacks(sound):
                assert sound == "quack"

        def a_bird():
            def it_flies(medium):
                assert medium == "air"

        def describe_birds():
            @fixture
            def medium():
                return "air"

            @behaves_like(a_duck, a_bird)
            def describe_something_that_quacks():
                @fixture
                def sound():
                    return "quack"

            @behaves_like(a_duck, a_bird)
            def describe_something_that_barks():
                @fixture
                def sound():
                    return "bark"
    """))

    result = testdir.runpytest()
    assert_outcomes(result, failed=1, passed=3)


def test_fixture(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        def a_duck():
            @fixture
            def sound():
                return "quack"

            def it_quacks(sound):
                assert sound == "quack"

        @behaves_like(a_duck)
        def describe_a_normal_duck():
            pass
    """))

    result = testdir.runpytest('-v')
    assert_outcomes(result, passed=1)


def test_override_fixture(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        def a_duck():
            @fixture
            def sound():
                return "quack"

            def it_quacks(sound):
                assert sound == "quack"

        @behaves_like(a_duck)
        def describe_something_that_barks():
            @fixture
            def sound():
                return "bark"
    """))

    result = testdir.runpytest('-v')
    assert_outcomes(result, failed=1)


def test_name_mangling(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        def thing():
            foo = 42
            def it_does_something():
                assert foo == 42

        @behaves_like(thing)
        def describe_something():
            foo = 4242
            def it_does_something():
                assert foo == 4242
    """))

    result = testdir.runpytest('-v')
    assert_outcomes(result, passed=2)


def test_nested_name_mangling(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        def thing():
            def it_does_something():
                pass
            def describe_thing():
                def it_does_something():
                    pass
                def describe_thing():
                    def it_does_something():
                        pass

        @behaves_like(thing)
        def describe_thing():
            def it_does_something():
                pass
            def describe_thing():
                def it_does_something():
                    pass
    """))

    result = testdir.runpytest('-v')
    assert_outcomes(result, passed=5)


def test_evaluated_once(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        from pytest import fixture
        from pytest_describe import behaves_like

        count = 0
        def thing():
            global count
            count += 1
            def is_evaluated_once():
                assert count == 1

        @behaves_like(thing)
        def describe_something():
            pass
        @behaves_like(thing)
        def describe_something_else():
            pass
    """))

    result = testdir.runpytest('-v')
    assert_outcomes(result, passed=2)



import py
import re

from util import assert_outcomes

pytest_plugins = 'pytester'


def test_collect(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        def describe_something():
            def is_foo():
                pass
            def can_bar():
                pass
            def _not_a_test():
                pass
        def describe_something_else():
            def describe_nested():
                def a_test():
                    pass
        def foo_not_collected():
            pass
        def test_something():
            pass
    """))

    result = testdir.runpytest('--collectonly')

    expected_regex = map(re.compile, [
        r"collected 4 item(s)?",
        r"\s*<DescribeBlock '?describe_something'?>",
        r"\s*<Function '?is_foo'?>",
        r"\s*<Function '?can_bar'?>",
        r"\s*<DescribeBlock '?describe_something_else'?>",
        r"\s*<DescribeBlock '?describe_nested'?>",
        r"\s*<Function '?a_test'?>",
        r"\s*<Function '?test_something'?>",
    ])
    for line in expected_regex:
        assert any([line.match(r) is not None for r in result.outlines])


def test_describe_evaluated_once(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_something.py').write(py.code.Source("""
        count = 0
        def describe_is_evaluated_only_once():
            global count
            count += 1
            def one():
                assert count == 1
            def two():
                assert count == 1
            def describe_nested():
                def three():
                    assert count == 1
    """))

    result = testdir.runpytest('-v')
    assert_outcomes(result, passed=3)



import py
from util import assert_outcomes

pytest_plugins = 'pytester'


def test_can_pass(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        def describe_something():
            def passes():
                assert True
            def describe_nested():
                def passes_too():
                    assert True
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=2)


def test_can_fail(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        def describe_something():
            def fails():
                assert False
            def describe_nested():
                def fails_too():
                    assert False
    """))

    result = testdir.runpytest()
    assert_outcomes(result, failed=2)


def test_can_fail_and_pass(testdir):
    a_dir = testdir.mkpydir('a_dir')
    a_dir.join('test_a.py').write(py.code.Source("""
        def describe_something():
            def describe_nested_ok():
                def passes():
                    assert True
            def describe_nested_bad():
                def fails():
                    assert False
    """))

    result = testdir.runpytest()
    assert_outcomes(result, passed=1, failed=1)
